#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from array import array
from ...common_descs import *
from supyr_struct.defs.tag_def import TagDef
from .objs.bitm import BitmTag

type_comment = """Type controls bitmap 'geometry'.
All dimensions must be a power-of-two except for SPRITES and INTERFACE:

*2D TEXTURES: Ordinary, 2D textures will be generated.

*3D TEXTURES: Volume textures will be generated from each
    sequence of 2D texture 'slices'.

*CUBE MAPS: Cube maps will be generated from each consecutive
    set of six 2D textures in each sequence. All faces
    of a cubemap must be square and have the same dimensions.

*SPRITES: Sprite texture pages will be generated.

*INTERFACE BITMAPS: Similar to 2D TEXTURES, but without mipmaps or
    the power-of-two restriction on their dimensions."""

format_comment = """Format controls how pixels will be stored internally.

*COMPRESSED WITH COLOR-KEY TRANSPARENCY: DXT1 compression(4-bits per pixel).
    For each 4x4 blocks of pixels, two colors are chosen that best represent
    the range of the colors in that block(c0 and c1). The colors are reduced
    to 16-bit and each of the 16 pixels is given a blending code(0 to 3).
    0 means the pixels color is c0 and 1 means its color is c1, with
    2 meaning to use 1/3(c1) + 2/3(c2), and 3 meaning 2/3(c1) + 1/3(c2).

    If an alpha exists, it is reduced to 1-bit(black or white).
    If a 4x4 block contains transparency, the blending codes are changed so
    that 2 means to use 1/2(c1) + 1/2(c2) and 3 means solid black(a=r=g=b=0).

*COMPRESSED WITH EXPLICIT ALPHA: DXT3 compression(8-bits per pixel).
    Same method as DXT1, but without the color-key transparency stuff.
    Alpha channel is quantized down to 4 bits per pixel(16 shades of gray).
    This format is best used where smooth gradients are not required in the
    alpha channel, and consistancy in shades between 4x4 chunks is important.

*COMPRESSED WITH INTERPOLATED ALPHA: DXT5 compression(8-bits per pixel).
    Same method as DXT1, but without the color-key transparency stuff.
    Alpha channel uses a method similar to DXT1. For each 4x4 block of
    pixels, two 8-bit shades of gray are chosen that best represent the
    range of values in that block(v0, v1). Each of the 16 pixels is given a
    blending code(0 to 7), with 0 meaning to use v0 and 1 meaning to use v1.
    The rest of the codes blend the v0 and v1 shades as shown:
        2 = (v0*6 + v1)/7      3 = (v0*5 + v1*2)/7
        4 = (v0*4 + v1*3)/7    5 = (v0*3 + v1*4)/7
        6 = (v0*2 + v1*5)/7    7 = (v0   + v1*6)/7
    If 100% white and 100% black are in the 4x4 block, these are used instead.
        2 = (v0*4 + v1)/5      3 = (v0*3 + v1*2)/5
        4 = (v0*2 + v1*3)/5    5 = (v0   + v1*4)/5
        6 = black              7 = white
    This allows very smooth gradients in the alpha, but if two neighboring
    4x4 blocks do not use the same v0 and v1 shades, it can be very noticible.

*16-BIT COLOR: Uses 16 bits per pixel. Depending on the alpha channel
    bitmaps are quantized to one of 3 different formats:
    r5g6b5(no alpha), a1r5g5b5(1-bit alpha), or a4r4g4b4(>1-bit alpha)

*32-BIT COLOR: Uses 32 bits per pixel. Very high quality and can have an alpha
    channel at no added cost. This format takes up the most memory, however.
    Bitmap formats are x8r8g8b8 and a8r8g8b8.

*MONOCHROME: Uses either 8 or 16 bits per pixel. This is an Xbox-only format.
    There are 4 different formats, each using an intensity and alpha channel.
    An intensity channel is essentially a monochrome rgb channel:
        a8: 8-bits per pixel. Intensity channel is solid black, with the
            pixel data being used for the alpha channel.
        y8: 8-bits per pixel. Alpha channel is solid white, with the pixel
            data being used for the intensity channel.
        ay8: 8-bits per pixel. Pixel data is used for both intensity and alpha.
        a8y8: 16-bits per pixel. Intensity and alpha channels each use their
            own separate pixel data, with 8 bits for each channel.

NOTE: Normal maps(a.k.a. bump maps) usually use 32-bit color.
    This is costly, and if there is no alpha you can usually use 16-bit
    r5g6b5 to save space without any noticible drop in quality ingame."""

usage_comment = """Usage controls how mipmaps are generated:

*ALPHA BLEND: Pixels with zero alpha are ignored in mipmaps, to prevent
    bleeding the transparent color.

*DEFAULT: Downsampling works normally, as in Photoshop.

*HEIGHT MAP: The bitmap is a height map, which will get converted to a bump map.
    Uses the 'bump height' below. Alpha is 1-bit. This is an Xbox-only format.

*DETAIL MAP: Mipmap color fades to gray and alpha fades to white.
    Color fading is controlled by the 'detail fade factor' below.

*LIGHT MAP: Generates no mipmaps. Do not use!

*VECTOR MAP: Used mostly for special effects; pixels are treated as XYZ vectors
    and are normalized after downsampling. Alpha is passed though unmodified."""

post_processing_comment = """
These properties control how mipmaps are processed."""

sprite_processing_comment = """
When creating a sprite group, specify the number and size of the textures
that the group is allowed to occupy. During importing, you will recieve
feedback about how well the alloted space was used."""

def get(): return bitm_def

def pixel_block_size(node, *a, **kwa):
    if isinstance(node, array):
        return node.itemsize*len(node)
    return len(node)

pixel_root = WhileArray('pixel_root',
    SUB_STRUCT=WhileArray('bitmap_pixels',
        SUB_STRUCT=UInt8Array('pixels', SIZE=pixel_block_size)
        )
    )

sprite = QStruct("sprite",
    SInt16("bitmap_index"),
    Pad(6),
    Float("left_side"),
    Float("right_side"),
    Float("top_side"),
    Float("bottom_side"),
    Float("registration_point_x"),
    Float("registration_point_y"),
    SIZE=32,
    )

sequence = Struct("sequence",
    ascii_str32("sequence_name"),
    SInt16("first_bitmap_index"),
    SInt16("bitmap_count"),
    Pad(16),
    reflexive("sprites", sprite, 64),
    SIZE=64,
    )

bitmap = Struct("bitmap",
    UEnum32('bitm_id', ('bitm', 'bitm'), DEFAULT='bitm', EDITABLE=False),
    UInt16("width", SIDETIP="pixels", EDITABLE=False),
    UInt16("height", SIDETIP="pixels", EDITABLE=False),
    UInt16("depth", SIDETIP="pixels", EDITABLE=False),
    SEnum16("type",
        "texture_2d",
        "texture_3d",
        "cubemap",
        "white",
        EDITABLE=False
        ),
    SEnum16("format",
        "a8",
        "y8",
        "ay8",
        "a8y8",
        #"-unused1-",
        #"-unused2-",
        ("r5g6b5", 6),
        #"-unused3-",
        ("a1r5g5b5", 8),
        ("a4r4g4b4", 9),
        ("x8r8g8b8", 10),
        ("a8r8g8b8", 11),
        #"-unused4-",
        #"-unused5-",
        ("dxt1", 14),
        ("dxt3", 15),
        ("dxt5", 16),
        ("p8_bump", 17),
        ),
    Bool16("flags",
        "power_of_2_dim",
        "compressed",
        "palletized",
        "swizzled",
        "linear",
        "v16u16",
        "unknown",
        "prefer_low_detail",
        "data_in_resource_map",
        ),
    UInt16("registration_point_x"),
    UInt16("registration_point_y"),
    UInt16("mipmaps"),
    FlUInt16("pixels", VISIBLE=False, EDITABLE=False),

    # this is the non-magic pointer into the map that the pixel data
    # is located at. if flags.data_in_resource_map is True and the
    # map is halo ce/pc/trial, the offset is into the bitmaps.map
    UInt32("pixels_offset", VISIBLE=False, EDITABLE=False),
    UInt32("pixels_meta_size", VISIBLE=False, EDITABLE=False),
    UInt32("bitmap_id_unknown1", VISIBLE=False, EDITABLE=False),
    UInt32("bitmap_data_pointer", VISIBLE=False, EDITABLE=False),
    UInt32("bitmap_id_unknown2", VISIBLE=False, EDITABLE=False),
    UInt32("base_address", VISIBLE=False, EDITABLE=False),
    SIZE=48,
    )

bitm_body = Struct("tagdata",
    SEnum16("type",
        "textures_2d",
        "textures_3d",
        "cubemaps",
        "sprites",
        "interface_bitmaps",
        COMMENT=type_comment
        ),
    SEnum16("format",
        "color_key_transparency",
        "explicit_alpha",
        "interpolated_alpha",
        "color_16bit",
        "color_32bit",
        "monochrome",
        COMMENT=format_comment
        ),
    SEnum16("usage",
        "alpha-blend",
        "default",
        "height_map",
        "detail_map",
        "light_map",
        "vector_map",
        COMMENT=usage_comment, DEFAULT=1
        ),
    Bool16("flags",
        "enable_diffusion_dithering",
        "disable_height_map_compression",
        "uniform_sprite_sequences",
        "sprite_bug_fix",
        ),
    QStruct("post_processing",
        float_zero_to_one("detail_fade_factor"),
        float_zero_to_one("sharpen_amount"),
        Float("bump_height", SIDETIP="repeats"),
        COMMENT=post_processing_comment
        ),
    Struct("sprite_processing",
        SEnum16("sprite_budget_size",
          {NAME: "x32",  VALUE: 0, GUI_NAME: "32x32"},
          {NAME: "x64",  VALUE: 1, GUI_NAME: "64x64"},
          {NAME: "x128", VALUE: 2, GUI_NAME: "128x128"},
          {NAME: "x256", VALUE: 3, GUI_NAME: "256x256"},
          {NAME: "x512", VALUE: 4, GUI_NAME: "512x512"},
          ),
        UInt16("sprite_budget_count"),
        COMMENT=sprite_processing_comment
        ),
    UInt16("color_plate_width",  SIDETIP="pixels", EDITABLE=False),
    UInt16("color_plate_height", SIDETIP="pixels", EDITABLE=False),

    rawdata_ref("compressed_color_plate_data", max_size=16777216),
    rawdata_ref("processed_pixel_data", max_size=16777216),

    Float("blur_filter_size", MIN=0.0, MAX=10.0, SIDETIP="[0,10] pixels"),
    float_neg_one_to_one("alpha_bias"),
    UInt16("mipmap_levels", MIN=0, SIDETIP="levels"),
    SEnum16("sprite_usage",
       "blend/add/subtract/max",
       "multiply/min",
       "double_multiply",
       ),
    UInt16("sprite_spacing", SIDETIP="pixels"),
    Pad(2),
    reflexive("sequences", sequence, 256,
        DYN_NAME_PATH='.sequence_name', IGNORE_SAFE_MODE=True),
    reflexive("bitmaps", bitmap, 2048, IGNORE_SAFE_MODE=True),
    SIZE=108, WIDGET=HaloBitmapTagFrame
    )

def get():
    return bitm_def

bitm_def = TagDef("bitm",
    blam_header('bitm', 7),
    bitm_body,

    ext=".bitmap", endian=">", tag_cls=BitmTag,
    subdefs = {'pixel_root':pixel_root}
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import array
import audioop
import re
import struct
import sys

from reclaimer.sounds import constants

BAD_PATH_CHAR_REMOVAL = re.compile(r'[<>:"|?*]{1, }')


def get_sample_chunk_size(compression, encoding, chunk_size=None):
    if chunk_size is None:
        chunk_size = constants.MAX_SAMPLE_CHUNK_SIZE

    if compression == constants.COMPRESSION_XBOX_ADPCM:
        chunk_size = (
            chunk_size // constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE
            ) * constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE
    elif compression == constants.COMPRESSION_IMA_ADPCM:
        chunk_size = (
            chunk_size // constants.IMA_ADPCM_DECOMPRESSED_BLOCKSIZE
            ) * constants.IMA_ADPCM_COMPRESSED_BLOCKSIZE

    return chunk_size - (
        chunk_size % get_block_size(compression, encoding))


def get_block_size(compression, encoding):
    if compression == constants.COMPRESSION_OGG:
        return 1

    if compression == constants.COMPRESSION_XBOX_ADPCM:
        block_size = constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE
    elif compression == constants.COMPRESSION_IMA_ADPCM:
        block_size = constants.IMA_ADPCM_COMPRESSED_BLOCKSIZE
    else:
        block_size = constants.sample_widths[compression]
    return block_size * constants.channel_counts.get(encoding, 1)


def get_samples_per_block(compression):
    if compression == constants.COMPRESSION_XBOX_ADPCM:
        return constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2
    elif compression == constants.COMPRESSION_IMA_ADPCM:
        return constants.IMA_ADPCM_DECOMPRESSED_BLOCKSIZE // 2
    return 1


def get_sample_count(sample_data, compression, encoding):
    if compression == constants.COMPRESSION_OGG:
        raise NotImplementedError(
            "Cannot yet calculate ogg sample count")

    block_size = get_block_size(compression, encoding)
    chunk_count = len(sample_data) // block_size

    if compression == constants.COMPRESSION_XBOX_ADPCM:
        chunk_count *= constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2
    elif compression == constants.COMPRESSION_IMA_ADPCM:
        chunk_count *= constants.IMA_ADPCM_DECOMPRESSED_BLOCKSIZE // 2

    return chunk_count


def byteswap_pcm16_sample_data(samples):
    return convert_pcm_to_pcm(
        samples,
        constants.COMPRESSION_PCM_16_LE,
        constants.COMPRESSION_PCM_16_BE)


def is_big_endian_pcm(compression):
    '''
    Returns True if the endianness of the compression modes are different.
    Only considers the linear PCM formats.
    '''
    assert compression in constants.PCM_FORMATS
    if constants.sample_widths.get(compression, 1) == 1:
        return False

    return bool(compression & 1)


def change_pcm_endianness(compression):
    '''
    Returns the compression mode with endianness opposite
    that of the provided compression mode.
    Only considers the linear PCM formats.
    '''
    assert compression in constants.PCM_FORMATS
    if constants.sample_widths.get(compression, 1) == 1:
        return compression

    return compression ^ 1


def change_pcm_width(compression, new_width):
    '''
    Returns the compression mode at the specified width, but
    with the endianness of the provided compression mode.
    Only considers the linear PCM formats.
    '''
    assert compression in constants.PCM_FORMATS
    is_big_endian = is_big_endian_pcm(compression)

    if new_width == 4:
        new_compression = constants.COMPRESSION_PCM_32_LE
    elif new_width == 3:
        new_compression = constants.COMPRESSION_PCM_24_LE
    elif new_width == 2:
        new_compression = constants.COMPRESSION_PCM_16_LE
    elif new_width == 1:
        new_compression = constants.COMPRESSION_PCM_8_SIGNED
        is_big_endian = False

    return new_compression | int(is_big_endian)


def convert_pcm_to_pcm(samples, compression, target_compression,
                       encoding=constants.ENCODING_MONO,
                       target_encoding=constants.ENCODING_MONO,
                       sample_rate=constants.SAMPLE_RATE_22K,
                       target_sample_rate=constants.SAMPLE_RATE_22K):
    '''
    Converts a stream of PCM audio to one with a different compression
    and/or encoding and/or sample rate.
    '''
    assert compression in constants.PCM_FORMATS
    assert target_compression in constants.PCM_FORMATS
    current_width = constants.sample_widths.get(compression, 1)
    target_width  = constants.sample_widths.get(target_compression, 1)
    if len(samples) % current_width:
        # ensure samples are a multiple of their width
        samples = samples[: len(samples) - (len(samples) % current_width)]

    if compression == constants.COMPRESSION_PCM_8_UNSIGNED:
        # bias by 128 to shift unsigned into signed
        samples = audioop.bias(samples, 1, 128)
    elif current_width > 1 and compression not in constants.NATIVE_ENDIANNESS_FORMATS:
        # byteswap samples to system endianness before processing
        samples = audioop.byteswap(samples, current_width)
        compression = change_pcm_endianness(compression)

    if current_width != target_width:
        samples = audioop.lin2lin(samples, current_width, target_width)

        # change compression to one with the correct sample width
        compression = change_pcm_width(compression, target_width)

    # make sure to convert to/from mono/stereo
    if encoding != target_encoding:
        if (encoding == constants.ENCODING_MONO and
            target_encoding == constants.ENCODING_STEREO):
            samples = audioop.tostereo(samples, target_width, 1, 1)
        elif (encoding == constants.ENCODING_STEREO and
              target_encoding == constants.ENCODING_MONO):
            samples = audioop.tomono(samples, target_width, 0.5, 0.5)
        else:
            raise ValueError("Can only convert between mono and stereo encodings.")
        encoding = target_encoding

    # convert sample rate if necessary
    if sample_rate != target_sample_rate:
        samples, _ = audioop.ratecv(
            samples, target_width, constants.channel_counts[encoding],
            sample_rate, target_sample_rate, None)
        sample_rate = target_sample_rate

    if target_compression == constants.COMPRESSION_PCM_8_UNSIGNED:
        # bias by 128 to shift signed back into unsigned
        samples = audioop.bias(samples, 1, 128)
    elif target_width > 1 and (is_big_endian_pcm(compression) !=
                               is_big_endian_pcm(target_compression)):
        # byteswap samples to target endianness
        samples = audioop.byteswap(samples, target_width)

    return samples


def convert_pcm_float32_to_pcm_32(sample_data):
    samples = array.array('f', sample_data)
    if sys.byteorder == "big":
        samples.byteswap()

    samples = [-0x7fFFffFF if val <= -1.0 else
               (0x7fFFffFF if val >=  1.0 else
                int(val * 0x7fFFffFF))
               for val in samples]

    return struct.pack("<%di" % len(samples), *samples)


def generate_mouth_data(sample_data, compression, sample_rate, encoding):
    assert compression in constants.PCM_FORMATS
    assert encoding in constants.channel_counts
    assert sample_rate > 0

    sample_width = constants.sample_widths[compression]
    channel_count = constants.channel_counts[encoding]

    if compression == constants.COMPRESSION_PCM_8_UNSIGNED:
        # bias by 128 to shift unsigned into signed
        sample_data = audioop.bias(sample_data, 1, 128)
    elif sample_width > 1 and compression not in constants.NATIVE_ENDIANNESS_FORMATS:
        # byteswap samples to system endianness before processing
        sample_data = audioop.byteswap(sample_data, sample_width)

    if sample_width == 2:
        sample_data = memoryview(sample_data).cast("h")
    elif sample_width == 4:
        sample_data = memoryview(sample_data).cast("i")

    # mouth data is sampled at 30Hz, so we divide the audio
    # sample_rate by that to determine how many samples we must
    # consider for each fragment. also, since mouth data doesn't
    # use multiple channels, and the audio samples are interleaved,
    # we multiply the channel count into the fragment_width.
    samples_per_mouth_sample = sample_rate / constants.SAMPLE_RATE_MOUTH_DATA
    fragment_width = int(
        channel_count * samples_per_mouth_sample + 0.5)

    # add fragment_width - 1 to round up to next multiple of fragment_width
    fragment_count = (
        len(sample_data) + fragment_width - 1) // fragment_width

    # used to scale the max fragment to the [0, 255] scale of a uint8
    scale_to_uint8 = 255 / ((1 << (sample_width * 8 - 1)) - 1)

    # generate mouth data samples
    mouth_data = bytearray(fragment_count)
    for i in range(fragment_count):
        fragment = sample_data[i * fragment_width:
                               (i + 1) * fragment_width]
        fragment_avg = sum(map(abs, fragment)) / samples_per_mouth_sample

        mouth_sample = fragment_avg * scale_to_uint8
        if mouth_sample >= 255:
            mouth_data[i] = 255
        else:
            mouth_data[i] = int(mouth_sample)

    # shift/scale the mouth samples based on the range of the mouth data
    mouth_avg = sum(mouth_data) / len(mouth_data)
    mouth_max = max(mouth_data)
    mouth_min = max(0, min(255, 2 * mouth_avg - mouth_max))

    mouth_range = (mouth_avg + mouth_max) / 2 - mouth_min
    if mouth_range == 0:
        # no range in the volume. don't try to scale
        # or shift, or else we'll divide by zero
        return bytes(mouth_data)

    for i in range(len(mouth_data)):
        mouth_sample = (mouth_data[i] - mouth_min) / mouth_range
        if mouth_sample >= 1.0:
            mouth_data[i] = 255
        elif mouth_sample <= 0.0:
            mouth_data[i] = 0
        else:
            mouth_data[i] = int(255 * mouth_sample)

    return bytes(mouth_data)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall-
# 	Labeled some reflexives and acouple values(Sep 26 2008)
# revision: 3		author: OrangeMohawk
# 	Position, Rotation, Scaling, Flashing/Fading, Bitmap Skewing.
# revision: 4		author: Lord Zedd
# 	Updated some things
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


chad_position_animation = Struct("animation",
    SInt32("frame_number"),
    QStruct("position", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=16
    )


chad_position = Struct("position",
    h3_reflexive("animation", chad_position_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_rotation_animation = Struct("animation",
    SInt32("frame_number"),
    ypr_float_rad("angle"),
    ENDIAN=">", SIZE=16
    )


chad_rotation = Struct("rotation",
    h3_reflexive("animation", chad_rotation_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_size_animation = Struct("animation",
    SInt32("frame_number"),
    QStruct("stretch", INCLUDE=xy_float),
    ENDIAN=">", SIZE=12
    )


chad_size = Struct("size",
    h3_reflexive("animation", chad_size_animation),
    h3_rawdata_ref("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


chad_color_animation = Struct("animation",
    SInt32("frame_number"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


chad_color = Struct("color",
    h3_reflexive("animation", chad_color_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_alpha_animation = Struct("animation",
    SInt32("frame_number"),
    Float("alpha"),
    ENDIAN=">", SIZE=8
    )


chad_alpha = Struct("alpha",
    h3_reflexive("animation", chad_alpha_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_alpha_unknown_animation = Struct("animation",
    SInt32("frame_number"),
    Float("alpha"),
    ENDIAN=">", SIZE=8
    )


chad_alpha_unknown = Struct("alpha_unknown",
    h3_reflexive("animation", chad_alpha_unknown_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_bitmap_animation = Struct("animation",
    SInt32("frame_number"),
    QStruct("movement_1", INCLUDE=xy_float),
    QStruct("movement_2", INCLUDE=xy_float),
    ENDIAN=">", SIZE=20
    )


chad_bitmap = Struct("bitmap",
    h3_reflexive("animation", chad_bitmap_animation),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=32
    )


chad_body = Struct("tagdata",
    Bool16("flags",
        ("loops", 1 << 1),
        ),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("position", chad_position),
    h3_reflexive("rotation", chad_rotation),
    h3_reflexive("size", chad_size),
    h3_reflexive("color", chad_color),
    h3_reflexive("alpha", chad_alpha),
    h3_reflexive("alpha_unknown", chad_alpha_unknown),
    h3_reflexive("bitmap", chad_bitmap),
    SInt32("number_of_frames", VISIBLE=False),
    ENDIAN=">", SIZE=92
    )


def get():
    return chad_def

chad_def = TagDef("chad",
    h3_blam_header('chad'),
    chad_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["chad"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Done
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

styl_combat_status_decay_options = (
    "latch_at_idle",
    "latch_at_alert",
    "latch_at_combat",
    )


styl_special_movement = Struct("special_movement",
    Bool32("special_movement_1",
        "jump",
        "climb",
        "vault",
        "mount",
        "hoist",
        "wall_jump",
        "n_a",
        ),
    ENDIAN=">", SIZE=4
    )


styl_behavior_list = Struct("behavior_list",
    ascii_str32("behavior_name"),
    ENDIAN=">", SIZE=32
    )


styl_body = Struct("tagdata",
    ascii_str32("name"),
    SEnum16("combat_status_decay_options", *styl_combat_status_decay_options),
    SInt16("unknown", VISIBLE=False),
    Bool32("style_control",
        "new_behaviors_default_to_on",
        ),
    Bool32("behaviors1",
        "general",
        "root",
        "null",
        "null_discrete",
        "obey",
        "guard",
        "follow_behavior",
        "ready",
        "smash_obstacle",
        "destroy_obstacle",
        "perch",
        "cover_friend",
        "blind_panic",
        "combat",
        "broken",
        "broken_behavior",
        "huddle_impulse",
        "huddle_behavior",
        "kamikaze_behavior",
        "broken_kamikaze_impulse",
        "broken_berserk_impulse",
        "broken_flee_impulse",
        "broken_scatter_impulse",
        "engage_0",
        "equipment",
        "engage_1",
        "fight",
        "melee_charge",
        "melee_leaping_charge",
        "surprise",
        "grenade_impulse",
        "anti_vehicle_grenade",
        ),
    Bool32("behaviors2",
        "stalk",
        "flank",
        "berserk_wander_impulse",
        "stalker_camo_control",
        "leader_abandoned_berserk",
        "unassailable_grenade_impulse",
        "perimeter",
        "perimeter_timeout_morph",
        "perimeter_infection_spew",
        "berserk",
        "shield_depleted_berserk",
        "last_man_berserk",
        "stuck_with_grenade_berserk",
        "presearch_0",
        "presearch_1",
        "presearch_uncover",
        "destroy_cover",
        "suppressing_fire",
        "grenade_uncover",
        "leap_on_cover",
        "leader",
        "search_sync",
        "engage_sync",
        "search_0",
        "search_1",
        "uncover",
        "investigate",
        "pursuit_sync",
        "pursuit",
        "refresh_target",
        "sense_target",
        "postsearch",
        ),
    Bool32("behaviors3",
        "coverme_investigate",
        "self_defense",
        "self_preservation",
        "cover",
        "cover_peek",
        "avoid",
        "evasion_impulse",
        "dive_impulse",
        "danger_cover_impulse",
        "danger_crouch_impulse",
        "proximity_melee",
        "proximity_self_preservation",
        "unreachable_enemy_cover",
        "unassailable_enemy_cover",
        "scary_target_cover",
        "group_emerge",
        "shield_depleted_cover",
        "retreat_0",
        "retreat_1",
        "retreat_grenade",
        "flee",
        "cower",
        "low_shield_retreat",
        "scary_target_retreat",
        "leader_dead_retreat",
        "peer_dead_retreat",
        "danger_retreat",
        "proximity_retreat",
        "charge_when_cornered",
        "surprise_retreat",
        "overheated_weapon_retreat",
        "ambush",
        ),
    Bool32("behaviors4",
        "ambush",
        "coordinated_ambush",
        "proximity_ambush",
        "vulnerable_enemy_ambush",
        "nowhere_to_run_ambush",
        "vehicle",
        "enter_vehicle",
        "enter_friendly_vehicle",
        "vehicle_enter_impulse",
        "vehicle_entry_engage_impulse",
        "vehicle_board",
        "vehicle_fight",
        "vehicle_fight_boost",
        "vehicle_charge",
        "vehicle_ram_behavior",
        "vehicle_cover",
        "damage_vehicle_cover",
        "exposed_rear_cover_impulse",
        "player_endagered_cover_impulse",
        "vehicle_avoid",
        "vehicle_pickup",
        "vehicle_player_pickup",
        "vehicle_exit_impulse",
        "danger_vehicle_exit_impulse",
        "vehicle_flip_impulse",
        "vehicle_flip",
        "vehicle_turtle",
        "vehicle_engage_patrol_impulse",
        "vehicle_engage_wander_impulse",
        "postcombat_0",
        "postcombat_1",
        "post_postcombat",
        ),
    Bool32("behaviors5",
        "check_friend",
        "shoot_corpse",
        "postcombat_approach",
        "alert_0",
        "alert_1",
        "idle_0",
        "idle_1",
        "wander_behavior",
        "flight_wander",
        "patrol",
        "fall_sleep",
        "buggers",
        "bugger_ground_uncover",
        "swarms",
        "swarm_root",
        "swarm_attack",
        "support_attack",
        "infect",
        "scatter",
        "combatforms",
        "combat_form_berserk_control",
        "eject_parasite",
        "sentinels",
        "enforcer_weapon_control",
        "grapple",
        "guardians",
        "guardian_surge",
        "guardian_proximity_surge",
        "guardian_danger_surge",
        "guardian_isolation_surge",
        "pureforms",
        "group_morph_impulse",
        ),
    Bool32("behaviors6",
        "arrival_morph_impulse",
        "pureform_default_impulse",
        "search_morph",
        "stealth_active_camo_control",
        "stealth_damage_morph",
        "stealth_stalk",
        "stealth_stalk_thwarted",
        "stealth_stalk_group",
        "stealth_charge_recover",
        "ranged_proximity_morph",
        "tank_distance_damage_morph",
        "tank_throttle_control",
        "stealth_morph",
        "tank_morph",
        "ranged_morph",
        "ranged_turtle",
        "ranged_uncover",
        "scarab",
        "scarab_root",
        "scarab_cure_isolation",
        "scarab_combat",
        "scarab_fight",
        "scarab_target_lock",
        "scarab_search",
        "scarab_search_pause",
        "atoms",
        "go_to",
        "activities",
        "activity",
        "posture",
        "activity_default",
        "special",
        ),
    Bool32("behaviors7",
        "formation",
        "grunt_scared_by_elite",
        "stunned",
        "cure_isolation",
        "deploy_turret",
        ),
    h3_reflexive("special_movement", styl_special_movement),
    h3_reflexive("behavior_list", styl_behavior_list),
    ENDIAN=">", SIZE=92
    )


def get():
    return styl_def

styl_def = TagDef("styl",
    h3_blam_header('styl'),
    styl_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["styl"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from copy import deepcopy
from reclaimer.animation import serialization
from reclaimer.animation import jma
from reclaimer.model import jms

__all__ = ("compress_animation", "decompress_animation")


def decompress_animation(anim, keep_compressed=True, endian=">"):
    if not anim.flags.compressed_data:
        return False

    jma_anim = jma.JmaAnimation(
        anim.name, anim.node_list_checksum, anim.type.enum_name,
        anim.frame_info_type.enum_name, anim.flags.world_relative
        )
    jma_anim.trans_flags_int = anim.trans_flags0 | (anim.trans_flags1 << 32)
    jma_anim.rot_flags_int   = anim.rot_flags0   | (anim.rot_flags1   << 32)
    jma_anim.scale_flags_int = anim.scale_flags0 | (anim.scale_flags1 << 32)

    # decompress the frame data
    _, jma_anim.frames = serialization.deserialize_compressed_frame_data(anim)

    # make some fake nodes so the serialization functions work
    jma_anim.nodes = jms.util.generate_fake_nodes(anim.node_count)

    # serialize the animation data
    def_data   = serialization.serialize_default_data(jma_anim, endian)
    frame_data = serialization.serialize_frame_data(jma_anim, endian)

    if keep_compressed:
        anim.offset_to_compressed_data = len(frame_data)
        frame_data += anim.frame_data.STEPTREE[anim.offset_to_compressed_data: ]
    else:
        anim.offset_to_compressed_data = 0
        anim.flags.compressed_data = False

    anim.frame_size = jma_anim.frame_data_frame_size
    anim.default_data.STEPTREE = def_data
    anim.frame_data.STEPTREE = frame_data
    return True


def compress_animation(anim, endian=">", **kw):
    jma_anim = kw.pop("jma_anim", None)
    if jma_anim is None:
        jma_anim = jma.JmaAnimation(
            anim.name, anim.node_list_checksum, anim.type.enum_name,
            anim.frame_info_type.enum_name, anim.flags.world_relative
            )
        jma_anim.trans_flags_int = anim.trans_flags0 | (anim.trans_flags1 << 32)
        jma_anim.rot_flags_int   = anim.rot_flags0   | (anim.rot_flags1   << 32)
        jma_anim.scale_flags_int = anim.scale_flags0 | (anim.scale_flags1 << 32)

        # create the node states from the frame_data and default_data
        jma_anim.frames = serialization.deserialize_frame_data(
            anim, None, True, endian)

        # make some fake nodes so the serialization functions work
        jma_anim.nodes = jms.util.generate_fake_nodes(anim.node_count)

        keyframes = kw.pop("keyframes", None)
        # either use provided keyframes, or generate them
        if keyframes is None:
            jma_anim.compress_animation(**kw)
        else:
            jma_anim.rot_keyframes   = keyframes[0]
            jma_anim.trans_keyframes = keyframes[1]
            jma_anim.scale_keyframes = keyframes[2]

    if jma_anim.root_node_info_applied:
        jma_anim = deepcopy(jma_anim)
        jma_anim.apply_root_node_info_to_states(True)

    # serialize the animation data
    comp_frame_data = serialization.serialize_compressed_frame_data(jma_anim)

    uncomp_frame_data = anim.frame_data.STEPTREE
    if anim.flags.compressed_data:
        del uncomp_frame_data[anim.offset_to_compressed_data: ]

    anim.offset_to_compressed_data = len(uncomp_frame_data)
    anim.flags.compressed_data = True
    anim.frame_data.STEPTREE = uncomp_frame_data + comp_frame_data

    return True



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

bloom_global = Struct("bloom_global",
    Bool16("shader_model",
        "enabled",
        "apply_after_hud",
        ),
    Pad(2),
    Float("bloom_size"),
    Float("bloom_exposure"),
    Float("bloom_mix_amount", MIN=0.0, MAX=1.0),
    QStruct("bloom_minimum_color", INCLUDE=rgb_float),
    QStruct("bloom_maximum_color", INCLUDE=rgb_float),
    SIZE=40,
    )

sppg_body = Struct("tagdata",
    Pad(4),
    reflexive("bloom_globals", bloom_global, 1),
    SIZE=244
    )

def get():
    return sppg_def

sppg_def = TagDef("sppg",
    blam_header_os('sppg'),
    sppg_body,

    ext=".shader_postprocess_globals", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.soso import *

specular_map_comment = """SPECULAR (COLOR) MAP
The specular color map is multiplied by the stock specular
to provide finer control over specular color.

*RGB holds the specular color.
*ALPHA can be used as an exponent mask to provide finer control
over the size of the specular highlights."""

normal_maps_comment = """NORMAL MAPS
*RED holds the horizontal normal data.
*GREEN holds the vertical normal data.
*BLUE holds the 'height', dimmer blue means dimmer specular and reflection.

*ALPHA is used for detail normal opacity:
A = no detail normal maps visible
B = detail normal 1 at full opacity
C = detail normal 2 at full opacity
D = also no detail normal maps visible

alpha brightness:
   A -------> B -------> C -------> D
  0/0.000   85/0.333  170/0.666  255/1.000

Normal coefficient is used to change the height/visiblity of the normal map."""

specular_tint_override_comment = """SPECULAR TINT OVERRIDE
These values override the specular tint colors in the stock part of the tag (the bottom).
Useful for if you want to have specific os and non-os tint values.

(Whetever these values are used is controlled by if the user has opensauce,
not by the opensauce settings.)"""

diffuse_lighting_comment = """DIFFUSE LIGHTING"""

specular_lighting_comment = """SPECULAR LIGHTING
Exponent controls the highlight size, the bigger the exponent, the smaller the highlight.
Coefficient controls the brightness of the highlights."""

os_reflection_prop_comment = """REFLECTION PROPERIES
When the opensauce extension is used the tint values in here are overwritten
by the ones in the os extension when the map is loaded."""

reflection = Struct("reflection",
    INCLUDE=reflection, COMMENT=os_reflection_prop_comment
    )

os_soso_ext = Struct("shader_model_extension",
    #Specular Color
    dependency_os("specular_color_map", "bitm", COMMENT=specular_map_comment),
    Float("specular_color_coefficient"),
    Pad(4),
    Float("specular_color_exponent"),
    Bool16("flags",
        "alpha_as_exponent_mask",
        ),

    Pad(2),
    #Base Normal Map
    dependency_os("base_normal_map", "bitm", COMMENT=normal_maps_comment),
    Float("base_normal_coefficient"),

    Pad(12),
    #Detail Normal Maps
    dependency_os("detail_normal_1_map", "bitm"),
    Float("detail_normal_1_coefficient"),
    Float("detail_normal_1_scale"),
    Float("detail_normal_1_v_scale"),

    Pad(4),
    dependency_os("detail_normal_2_map", "bitm"),
    Float("detail_normal_2_coefficient"),
    Float("detail_normal_2_scale"),
    Float("detail_normal_2_v_scale"),

    Pad(4),
    #Specular Tint Override
    float_zero_to_one("perpendicular_brightness", COMMENT=specular_tint_override_comment),
    QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
    float_zero_to_one("parallel_brightness"),
    QStruct("parallel_tint_color", INCLUDE=rgb_float),

    Bool16("diffuse_lighting_flags",
        { NAME: "do_not_use_dlms", GUI_NAME: "do not use dlms (bsp)" },
        COMMENT=diffuse_lighting_comment
        ),
    Pad(4),
    Pad(2),

    Float("specular_lighting_exponent", COMMENT=specular_lighting_comment),
    Float("specular_lighting_coefficient"),
    SIZE=192,
    )

soso_attrs = Struct("soso_attrs",
    #Model Shader Properties
    model_shader,

    Pad(16),
    #Color-Change
    SEnum16("color_change_source", *function_names, COMMENT=cc_comment),

    Pad(30),
    #Self-Illumination
    self_illumination,

    Pad(12),
    #Diffuse, Multipurpose, and Detail Maps
    maps,

    reflexive("os_shader_model_ext", os_soso_ext, 1),

    #Texture Scrolling Animation
    texture_scrolling,

    Pad(8),
    #Reflection Properties
    reflection,
    Pad(16),

    FlFloat("unknown0", VISIBLE=False),
    BytesRaw("unknown1", SIZE=16, VISIBLE=False),  # little endian dependency
    SIZE=400
    )

soso_body = Struct("tagdata",
    shdr_attrs,
    soso_attrs
    )

def get():
    return soso_def

soso_def = TagDef("soso",
    blam_header('soso', 2),
    soso_body,

    ext=".shader_model", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Labelled Raw Identifier
# revision: 3		author: Lord Zedd
# 	Standardized model data and updated
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sLdT_lightmap_instanced_meshe = Struct("instanced_meshe",
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    SInt32("unknown_index"),
    ENDIAN=">", SIZE=16
    )


sLdT_lightmap_unknown_1 = Struct("unknown_1",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


sLdT_lightmap_instanced_geometry = Struct("instanced_geometry",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("instanced_mesh_index"),
    SInt16("unknown_b_index", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


sLdT_lightmap_unknown_b = Struct("unknown_b",
    BytesRaw("unknown", SIZE=72, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=72
    )


sLdT_lightmap_meshe_part = Struct("part",
    SInt16("material_index"),
    SInt16("unknown_nodey_index", VISIBLE=False),
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("subpart_index"),
    SInt16("subpart_count"),
    SInt8("unknown_enum", VISIBLE=False),
    Bool8("flags",
        ("water", 1 << 3),
        ),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=16
    )


sLdT_lightmap_meshe_subpart = Struct("subpart",
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("part_index"),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=8
    )


sLdT_lightmap_meshe_instanced_geometry_indice_instanced_geometry_mesh_content = Struct("instanced_geometry_mesh_content",
    SInt16("instanced_geometry_index"),
    ENDIAN=">", SIZE=2
    )


sLdT_lightmap_meshe_instanced_geometry_indice = Struct("instanced_geometry_indice",
    SInt16("instanced_geometry_mesh_index_1"),
    SInt16("instanced_geometry_mesh_index_2"),
    h3_reflexive("instanced_geometry_mesh_contents", sLdT_lightmap_meshe_instanced_geometry_indice_instanced_geometry_mesh_content),
    ENDIAN=">", SIZE=16
    )


sLdT_lightmap_meshe_unknown_water = Struct("unknown_water",
    SInt16("unknown"),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sLdT_lightmap_meshe = Struct("meshe",
    h3_reflexive("parts", sLdT_lightmap_meshe_part),
    h3_reflexive("subparts", sLdT_lightmap_meshe_subpart),
    Array("vertex_buffer_index_array", SUB_STRUCT=SInt16("vertex_buffer_index"), SIZE=8),
    SInt16("index_buffer_index_1"),
    SInt16("index_buffer_index_2"),
    Bool8("flags",
        "has_vertex_colors",
        ),
    SInt8("rigid_node"),
    SEnum8("vertex_type", *pmdf_meshe_vertex_type),
    SEnum8("prt_type", *pmdf_meshe_prt_type),
    SEnum8("index_buffer_type", *pmdf_meshe_index_buffer_type),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_reflexive("instanced_geometry_indices", sLdT_lightmap_meshe_instanced_geometry_indice),
    h3_reflexive("unknown_water", sLdT_lightmap_meshe_unknown_water),
    ENDIAN=">", SIZE=76
    )


sLdT_lightmap_compression_info = Struct("compression_info",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    QStruct("position_bounds_x", INCLUDE=from_to),
    QStruct("position_bounds_y", INCLUDE=from_to),
    QStruct("position_bounds_z", INCLUDE=from_to),
    QStruct("texcoord_bounds_x", INCLUDE=from_to),
    QStruct("texcoord_bounds_y", INCLUDE=from_to),
    ENDIAN=">", SIZE=44
    )


sLdT_lightmap_unknown_nodey = Struct("unknown_nodey",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Array("node_index_array", SUB_STRUCT=SInt8("node_index"), SIZE=4, VISIBLE=False),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sLdT_lightmap_unknown_3 = Struct("unknown_3",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_rawdata_ref("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


sLdT_lightmap_unknown_meshe_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


sLdT_lightmap_unknown_meshe = Struct("unknown_meshe",
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    h3_reflexive("unknown_1", sLdT_lightmap_unknown_meshe_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


sLdT_lightmap_node_map_unknown = Struct("unknown",
    UInt8("node_index"),
    ENDIAN=">", SIZE=1
    )


sLdT_lightmap_node_map = Struct("node_map",
    h3_reflexive("unknown", sLdT_lightmap_node_map_unknown),
    ENDIAN=">", SIZE=12
    )


sLdT_lightmap_unknown_5_unknown = Struct("unknown",
    BytesRaw("unknown", SIZE=48, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


sLdT_lightmap_unknown_5 = Struct("unknown_5",
    h3_reflexive("unknown", sLdT_lightmap_unknown_5_unknown),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


sLdT_lightmap_unknown_yo = Struct("unknown_yo",
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt16("unknown_index"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


sLdT_lightmap = Struct("lightmap",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("bsp_index"),
    SInt32("structure_checksum"),
    Pad(216),
    h3_dependency("primary_map"),
    h3_dependency("intensity_map"),
    h3_reflexive("instanced_meshes", sLdT_lightmap_instanced_meshe),
    h3_reflexive("unknown_1", sLdT_lightmap_unknown_1),
    h3_reflexive("instanced_geometry", sLdT_lightmap_instanced_geometry),
    h3_reflexive("unknown_b", sLdT_lightmap_unknown_b),
    SInt32("unknown_2", VISIBLE=False),
    h3_reflexive("meshes", sLdT_lightmap_meshe),
    h3_reflexive("compression_info", sLdT_lightmap_compression_info),
    h3_reflexive("unknown_nodey", sLdT_lightmap_unknown_nodey),
    h3_reflexive("unknown_3", sLdT_lightmap_unknown_3),
    BytesRaw("unknown_4", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_meshes", sLdT_lightmap_unknown_meshe),
    h3_reflexive("node_maps", sLdT_lightmap_node_map),
    h3_reflexive("unknown_5", sLdT_lightmap_unknown_5),
    BytesRaw("unknown_6", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_yo", sLdT_lightmap_unknown_yo),
    zone_asset("asset_data"),
    ENDIAN=">", SIZE=436
    )


sLdT_airprobe = Struct("airprobe",
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    h3_string_id("unknown_3", VISIBLE=False),
    Bool16("unknown_4", *unknown_flags_16),
    SInt16("unknown_5", VISIBLE=False),
    BytesRaw("unknown_6", SIZE=72, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=92
    )


sLdT_unknown_2 = Struct("unknown_2",
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=4, VISIBLE=False),
    BytesRaw("unknown", SIZE=72, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=80
    )


sLdT_unknown_3_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown", SIZE=84, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=84
    )


sLdT_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown_0", SIZE=32, VISIBLE=False),
    h3_reflexive("unknown_1", sLdT_unknown_3_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=44
    )


sLdT_body = Struct("tagdata",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("lightmap", sLdT_lightmap),
    BytesRaw("unknown_1", SIZE=24, VISIBLE=False),
    h3_reflexive("airprobes", sLdT_airprobe),
    h3_reflexive("unknown_2", sLdT_unknown_2),
    h3_reflexive("unknown_3", sLdT_unknown_3),
    BytesRaw("unknown_4", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=88
    )


def get():
    return sLdT_def

sLdT_def = TagDef("sLdT",
    h3_blam_header('sLdT'),
    sLdT_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sLdT"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.util.matrices import Matrix, quaternion_to_matrix


__all__ = ("compile_physics", )


def compile_physics(phys_tag, jms_model_markers, updating=True):
    tagdata = phys_tag.data.tagdata
    mass_points = tagdata.mass_points.STEPTREE

    if not updating:
        # making fresh physics tag. use default values
        tagdata.radius = -1.0
        tagdata.moment_scale = 0.3
        tagdata.mass = 1.0
        tagdata.density = 1.0
        tagdata.gravity_scale = 1.0
        tagdata.ground_friction = 0.2
        tagdata.ground_depth = 0.2
        tagdata.ground_damp_fraction = 0.05
        tagdata.ground_normal_k1 = 0.7071068
        tagdata.ground_normal_k0 = 0.5
        tagdata.water_friction = 0.05
        tagdata.water_depth = 0.25
        tagdata.water_density = 1.0
        tagdata.air_friction = 0.001

    existing_mp_names = {}
    for i in range(len(mass_points)):
        existing_mp_names[mass_points[i].name.lower()] = i

    mass_points_to_update = {}
    for marker in jms_model_markers:
        name = marker.name.lower()
        if name in existing_mp_names:
            mass_points_to_update[name] = mass_points[existing_mp_names[name]]

    del mass_points[:]

    # update the mass points and/or make new ones
    for marker in jms_model_markers:
        rotation = quaternion_to_matrix(
            marker.rot_i, marker.rot_j,
            marker.rot_k, marker.rot_w)

        name = marker.name
        if name in existing_mp_names:
            mass_points.append(mass_points_to_update[name])
        else:
            mass_points.append()

        mp = mass_points[-1]
        if name not in existing_mp_names:
            # set default values
            mp.relative_mass = 1.0
            mp.relative_density = 1.0
            mp.friction_parallel_scale = 1.0
            mp.friction_perpendicular_scale = 1.0

        forward = rotation * Matrix(((1, ), (0, ), (0, )))
        up      = rotation * Matrix(((0, ), (0, ), (1, )))
        mp.up[:]       = up[0][0],      up[1][0],      up[2][0]
        mp.forward[:]  = forward[0][0], forward[1][0], forward[2][0]
        mp.position[:] = marker.pos_x/100, marker.pos_y/100, marker.pos_z/100
        mp.name = name
        mp.radius = marker.radius/100
        mp.model_node = marker.parent

    phys_tag.calc_internal_data()



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from copy import copy, deepcopy

try:
    from mozzarilla.widgets.field_widgets import ReflexiveFrame, HaloRawdataFrame,\
         TextFrame, ColorPickerFrame, EntryFrame, SoundSampleFrame,\
         DynamicArrayFrame, Halo2BitmapTagFrame
except Exception:
    ReflexiveFrame = HaloRawdataFrame = TextFrame = ColorPickerFrame =\
                     EntryFrame = SoundSampleFrame = DynamicArrayFrame =\
                     Halo2BitmapTagFrame = None
from reclaimer.common_descs import *
from reclaimer.h2.constants import STEPTREE, DYN_NAME_PATH, NAME_MAP,\
     COMMENT, TOOLTIP, WIDGET, MAX, MAX_REFLEXIVE_COUNT, VISIBLE, ORIENT,\
     MAX_TAG_PATH_LEN, DEFAULT, h2_tag_class_fcc_to_ext
from reclaimer.h2.field_types import *
from reclaimer.h2.enums import *

from supyr_struct.defs.block_def import BlockDef


resource = Struct("resource",
    SEnum8("type",
        "tag_block",  # reflexive
        "tag_data",   # rawdata
        "vertex_buffer",
        ),
    SInt8("unknown0", DEFAULT=2, VISIBLE=False),
    SInt16("primary_locator2", VISIBLE=False),
    SInt16("primary_locator",
        TOOLTIP="offset into the block section that the tag block/\n"
                "tag data/vertex buffer header block is located at."),
    SInt16("secondary_locator",
        TOOLTIP="tag block size in bytes(-1 if not a tag block)"),
    UInt32("resource_data_size",
        TOOLTIP="size of the rsrc section in bytes"),
    UInt32("resource_data_offset",
        TOOLTIP="offset into the block section that the rsrc section\n"
                "is located at."),
    SIZE=16
    )

def h2_tag_class(*args, **kwargs):
    '''
    A macro for creating a tag_class enum desc with the
    enumerations set to the provided tag_class fcc's.
    '''
    kwargs["class_mapping"] = h2_tag_class_fcc_to_ext
    return tag_class(*args, **kwargs)

def h2_reflexive(name, substruct, max_count=MAX_REFLEXIVE_COUNT, *names, **desc):
    '''This function serves to macro the creation of a reflexive'''
    desc.update(
        INCLUDE=h2_reflexive_struct,
        STEPTREE=Array(name + " array",
            SIZE=".size", MAX=max_count,
            SUB_STRUCT=substruct, WIDGET=ReflexiveFrame
            ),
        SIZE=8
        )
    if DYN_NAME_PATH in desc:
        desc[STEPTREE][DYN_NAME_PATH] = desc.pop(DYN_NAME_PATH)
    if names:
        name_map = {}
        for i in range(len(names)):
            e_name = BlockDef.str_to_name(None, names[i])
            name_map[e_name] = i

        desc[STEPTREE][NAME_MAP] = name_map

    return H2Reflexive(name, **desc)


def h2_rawtext_ref(name, f_type=StrRawLatin1, max_size=None,
                   widget=TextFrame, **kwargs):
    '''This function serves to macro the creation of a rawdata reference'''
    ref_struct = dict(h2_rawdata_ref_struct)
    if COMMENT in kwargs: ref_struct[COMMENT] = kwargs.pop(COMMENT)
    if TOOLTIP in kwargs: ref_struct[TOOLTIP] = kwargs.pop(TOOLTIP)
    kwargs.update(WIDGET=widget)
    ref_struct[0] = dict(ref_struct[0])
    ref_struct[0][VISIBLE] = VISIBILITY_METADATA
    if max_size is not None:
        ref_struct[0][MAX] = max_size
        kwargs[MAX] = max_size

    return H2RawdataRef(name,
        INCLUDE=ref_struct, ORIENT="v",
        STEPTREE=f_type("data",
            SIZE=".size", GUI_NAME=name.replace('_', ' '), **kwargs)
        )

def h2_rawdata_ref(name, f_type=BytearrayRaw, max_size=None,
                   widget=HaloRawdataFrame, **kwargs):
    '''This function serves to macro the creation of a rawdata reference'''
    ref_struct = dict(h2_rawdata_ref_struct)
    if COMMENT in kwargs: ref_struct[COMMENT] = kwargs.pop(COMMENT)
    if TOOLTIP in kwargs: ref_struct[TOOLTIP] = kwargs.pop(TOOLTIP)
    if max_size is not None:
        ref_struct[0] = dict(ref_struct[0])
        ref_struct[0][MAX] = max_size
        kwargs[MAX] = max_size

    if widget is not None:
        kwargs[WIDGET] = widget

    return H2RawdataRef(name,
        INCLUDE=ref_struct,
        STEPTREE=f_type("data", GUI_NAME="", SIZE=".size", **kwargs)
        )


def h2_dependency(name='tag ref', valid_ids=None, **kwargs):
    '''This function serves to macro the creation of a tag dependency'''
    if isinstance(valid_ids, tuple):
        valid_ids = h2_tag_class(*valid_ids)
    elif isinstance(valid_ids, str):
        valid_ids = h2_tag_class(valid_ids)
    elif valid_ids is None:
        valid_ids = valid_h2_tags

    kwargs.setdefault(ORIENT, "h")

    return H2TagRef(name,
        valid_ids,
        UInt32("id", VISIBLE=VISIBILITY_METADATA),
        SInt32("path length",
            MAX=MAX_TAG_PATH_LEN, OFFSET=4, VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
        STEPTREE=StrTagRef(
            "filepath", SIZE=tag_ref_str_size, GUI_NAME="", MAX=254),
        **kwargs
        )


def h2_blam_header(tagid, version=1):
    '''This function serves to macro the creation of a tag header'''
    header_desc = dict(h2_tag_header)
    header_desc[1] = dict(header_desc[1])
    header_desc[5] = dict(header_desc[5])
    header_desc[1][DEFAULT] = tagid
    header_desc[5][DEFAULT] = version
    return header_desc


valid_h2_tags = h2_tag_class(*h2_tag_class_fcc_to_ext.keys())


# ###########################################################################
# The order of element in all the enumerators is important(DONT SHUFFLE THEM)
# ###########################################################################

# DO NOT MODIFY ANY OF THESE SCRIPT ENUMS.
# The exact wording is important as the script extractor uses
# these strings to reconstruct scripts
script_types = (
    "startup",
    "dormant",
    "continuous",
    "static",
    "stub",
    "command_script",
    )
script_object_types = (
    "unparsed",
    "special form",
    "function name",
    "passthrough",
    "void",
    "boolean",
    "short",
    "long",
    "string_id",
    "unit_seat_mapping",

    "trigger_volume",
    "cutscene_flag",
    "cutscene_camera_point",
    "cutscene_title",
    "cutscene_recording",

    "device_group",
    "ai_command_list",
    "ai_command_script",
    "ai_behavior",
    "ai_orders",

# DO NOT MODIFY ANY OF THESE SCRIPT ENUMS.
# The exact wording is important as the script extractor uses
# these strings to reconstruct scripts

    "starting_profile",

    "conversation",
    "structure_bsp",
    "navpoint",
    "point reference",
    "hud_message",
    "object_list",

    "effect",
    "damage",
    "looping_sound",
    "animation_graph",
    "damage_effect",

    "object_definition",
    "shader",
    "render model",
    "structure definition",
    "lightmap definition",
    "game_difficulty",
    "actor_type",
    "hud_corner",

    "model_state",
    "unit",
    "object_name",
    "unit_name",
    "vehicle_name",
    "weapon_name",
    "device_name",
    "scenery_name",
    )

#Shared Enumerator options
old_materials_list = materials_list
# old materials list that is referenced in tags, but seems to be unused
del materials_list


# Descriptors
h2_blam_engine_id = dict(blam_engine_id)
h2_blam_engine_id[DEFAULT] = "b2am"
h2_tag_header = Struct("blam_header",
    Pad(36),
    UEnum32("tag class",
        GUI_NAME="tag class", INCLUDE=valid_h2_tags, EDITABLE=False
        ),
    UInt32("checksum", DEFAULT=0x4D6F7A7A, EDITABLE=False),
    UInt32("header size", DEFAULT=64, EDITABLE=False),
    Bool64("flags",
        "edited with mozz",
        EDITABLE=False
        ),
    UInt16("version", DEFAULT=1, EDITABLE=False),
    UInt8("integrity0", DEFAULT=0, EDITABLE=False),
    UInt8("integrity1", DEFAULT=255, EDITABLE=False),
    h2_blam_engine_id,
    VISIBLE=VISIBILITY_METADATA, SIZE=64
    )


h2_reflexive_struct = H2Reflexive('reflexive',
    SInt32("size", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False, DEFAULT=0xFFFFFFFF),
    )

h2_rawdata_ref_struct = H2RawdataRef('rawdata_ref',
    SInt32("size", GUI_NAME="", SIDETIP="bytes", EDITABLE=False),
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False, DEFAULT=0xFFFFFFFF),
    ORIENT='h'
    )

def h2_string_id(name, **kwargs):
    return string_id(name, 16, 8, 8, **kwargs)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

scale_comment = """DETAIL SOUND PERIOD SCALING
As the sound's input scale changes from zero to one, these modifiers move between
the two values specified here. The sound will play using the current scale modifier
multiplied by the value specified below. (0 values are ignored.)"""

random_spatialisation_comment = """RANDOM SPATIALIZATION
If the sound specified above is not stereo it will be randomly spatialized according
to the following contraints. If both lower and upper bounds are zero for any of
the following fields, the sound's position will be randomly selected from
all the possible directions or distances."""

detail_sound = Struct("detail_sound",
    dependency("sound", "snd!"),
    from_to_sec('random_period_bounds'),
    Float("gain"),
    Bool32("flags",
        "dont_play_with_alternate",
        "dont_play_without_alternate",
        ),

    Pad(48),
    from_to_rad('yaw_bounds', COMMENT=random_spatialisation_comment),  # radians
    from_to_rad('pitch_bounds'),  # radians
    from_to_wu('distance_bounds'),  # world units

    SIZE=104
    )

track = Struct("track",
    Bool32("flags",
        "fade_in_at_start",
        "fade_out_at_stop",
        "fade_in_alternate",
        ),
    Float("gain"),
    float_sec("fade_in_duration"),
    float_sec("fade_out_duration"),

    Pad(32),
    dependency("start", "snd!"),
    dependency("loop", "snd!"),
    dependency("end", "snd!"),

    Pad(32),
    dependency("alternate_loop", "snd!"),
    dependency("alternate_end", "snd!"),
    SIZE=160
    )


lsnd_body = Struct("tagdata",
    Bool32("flags",
        "deafening_to_ai",
        "not_a_loop",
        "stops_music",
        ),
    Float("detail_sound_period_at_zero", COMMENT=scale_comment),
    FlFloat("unknown0", DEFAULT=1.0, VISIBLE=False),
    FlFloat("unknown1", DEFAULT=1.0, VISIBLE=False),
    Float("detail_sound_period_at_one"),
    FlFloat("unknown2", DEFAULT=1.0, VISIBLE=False),
    FlFloat("unknown3", DEFAULT=1.0, VISIBLE=False),
    FlSInt16("unknown4", DEFAULT=-1, VISIBLE=False),
    FlSInt16("unknown5", DEFAULT=-1, VISIBLE=False),
    FlFloat("unknown6", DEFAULT=1.0, VISIBLE=False),
    Pad(8),
    dependency("continuous_damage_effect", "cdmg"),

    reflexive("tracks", track, 4),
    reflexive("detail_sounds", detail_sound, 32,
        DYN_NAME_PATH='.sound.filepath'),

    SIZE=84,
    )


def get():
    return lsnd_def

lsnd_def = TagDef("lsnd",
    blam_header("lsnd", 3),
    lsnd_body,

    ext=".sound_looping", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .file import *
from .marker import *
from .material import *
from .merged_model import *
from .merged_region import *
from .mesh import *
from .model import *
from .node import *
from .triangle import *
from .vertex import *
from . import util



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: bfixer117
# 	Named some more nodes.
# revision: 3		author: Lord Zedd
# 	Standardized model data and updated
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


mode_region_permutation = Struct("permutation",
    h3_string_id("name"),
    SInt16("mesh_index"),
    SInt16("mesh_count"),
    SInt32("unknown_0", VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


mode_region = Struct("region",
    h3_string_id("name"),
    h3_reflexive("permutations", mode_region_permutation),
    ENDIAN=">", SIZE=16
    )


mode_instance = Struct("instance",
    h3_string_id("name"),
    SInt32("node_index"),
    Float("default_scale"),
    QStruct("inverse_forward", INCLUDE=ijk_float),
    QStruct("inverse_left", INCLUDE=ijk_float),
    QStruct("inverse_up", INCLUDE=ijk_float),
    QStruct("inverse_position", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=60
    )


mode_node = Struct("node",
    h3_string_id("name"),
    SInt16("parent_node"),
    SInt16("first_child_node"),
    SInt16("next_sibling_node"),
    SInt16("import_node_index", VISIBLE=False),
    QStruct("default_translation", INCLUDE=xyz_float),
    QStruct("default_rotation", INCLUDE=ijkw_float),
    Float("default_scale"),
    QStruct("inverse_forward", INCLUDE=ijk_float),
    QStruct("inverse_left", INCLUDE=ijk_float),
    QStruct("inverse_up", INCLUDE=ijk_float),
    QStruct("inverse_position", INCLUDE=xyz_float),
    Float("distance_from_parent"),
    ENDIAN=">", SIZE=96
    )


mode_marker_group_marker = Struct("marker",
    SInt8("region_index"),
    SInt8("permutation_index"),
    SInt8("node_index"),
    SInt8("unknown", VISIBLE=False),
    QStruct("translation", INCLUDE=xyz_float),
    QStruct("rotation", INCLUDE=ijkw_float),
    Float("scale"),
    ENDIAN=">", SIZE=36
    )


mode_marker_group = Struct("marker_group",
    h3_string_id("name"),
    h3_reflexive("markers", mode_marker_group_marker),
    ENDIAN=">", SIZE=16
    )


mode_material_propertie = Struct("propertie",
    SInt16("type"),
    SInt16("int_value"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    Float("real_value"),
    ENDIAN=">", SIZE=12
    )


mode_material = Struct("material",
    h3_dependency("shader"),
    h3_reflexive("properties", mode_material_propertie),
    SInt32("unknown_0", VISIBLE=False),
    SInt8("breakable_surface_index"),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    SInt8("unknown_3", VISIBLE=False),
    ENDIAN=">", SIZE=36
    )


mode_meshe_part = Struct("part",
    SInt16("material_index"),
    SInt16("unknown_nodey_index", VISIBLE=False),
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("subpart_index"),
    SInt16("subpart_count"),
    SInt8("unknown_enum", VISIBLE=False),
    Bool8("flags",
        ("water", 1 << 3),
        ),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=16
    )


mode_meshe_subpart = Struct("subpart",
    SInt16("index_buffer_start"),
    SInt16("index_buffer_count"),
    SInt16("part_index"),
    SInt16("vertex_count"),
    ENDIAN=">", SIZE=8
    )


mode_meshe_instanced_geometry_indice_instanced_geometry_mesh_content = Struct("instanced_geometry_mesh_content",
    SInt16("instanced_geometry_index"),
    ENDIAN=">", SIZE=2
    )


mode_meshe_instanced_geometry_indice = Struct("instanced_geometry_indice",
    SInt16("instanced_geometry_mesh_index_1"),
    SInt16("instanced_geometry_mesh_index_2"),
    h3_reflexive("instanced_geometry_mesh_contents", mode_meshe_instanced_geometry_indice_instanced_geometry_mesh_content),
    ENDIAN=">", SIZE=16
    )


mode_meshe_unknown_water = Struct("unknown_water",
    SInt16("unknown"),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


mode_meshe = Struct("meshe",
    h3_reflexive("parts", mode_meshe_part),
    h3_reflexive("subparts", mode_meshe_subpart),
    Array("vertex_buffer_index_array", SUB_STRUCT=SInt16("vertex_buffer_index"), SIZE=8),
    SInt16("index_buffer_index_1"),
    SInt16("index_buffer_index_2"),
    Bool8("flags",
        "has_vertex_colors",
        ),
    SInt8("rigid_node"),
    SEnum8("vertex_type", *pmdf_meshe_vertex_type),
    SEnum8("prt_type", *pmdf_meshe_prt_type),
    SEnum8("index_buffer_type", *pmdf_meshe_index_buffer_type),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    h3_reflexive("instanced_geometry_indices", mode_meshe_instanced_geometry_indice),
    h3_reflexive("unknown_water", mode_meshe_unknown_water),
    ENDIAN=">", SIZE=76
    )


mode_compression_info = Struct("compression_info",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    QStruct("position_bounds_x", INCLUDE=from_to),
    QStruct("position_bounds_y", INCLUDE=from_to),
    QStruct("position_bounds_z", INCLUDE=from_to),
    QStruct("texcoord_bounds_x", INCLUDE=from_to),
    QStruct("texcoord_bounds_y", INCLUDE=from_to),
    ENDIAN=">", SIZE=44
    )


mode_unknown_nodey = Struct("unknown_nodey",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    Array("node_index_array", SUB_STRUCT=SInt8("node_index"), SIZE=4, VISIBLE=False),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


mode_unknown_5 = Struct("unknown_5",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_rawdata_ref("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


mode_unknown_meshe_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


mode_unknown_meshe = Struct("unknown_meshe",
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    h3_reflexive("unknown_1", mode_unknown_meshe_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


mode_node_map_unknown = Struct("unknown",
    UInt8("node_index"),
    ENDIAN=">", SIZE=1
    )


mode_node_map = Struct("node_map",
    h3_reflexive("unknown", mode_node_map_unknown),
    ENDIAN=">", SIZE=12
    )


mode_unknown_7_unknown = Struct("unknown",
    BytesRaw("unknown", SIZE=48, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


mode_unknown_7 = Struct("unknown_7",
    h3_reflexive("unknown", mode_unknown_7_unknown),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


mode_unknown_yo = Struct("unknown_yo",
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt16("unknown_index"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


mode_unknown_9 = Struct("unknown_9",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=7, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=28
    )


mode_unknown_10 = Struct("unknown_10",
    BytesRaw("unknown", SIZE=336, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=336
    )


mode_runtime_node = Struct("runtime_node",
    QStruct("default_rotation", INCLUDE=ijkw_float),
    QStruct("default_translation", INCLUDE=xyz_float),
    Float("default_scale"),
    ENDIAN=">", SIZE=32
    )


mode_body = Struct("tagdata",
    h3_string_id("name"),
    Bool16("flags",
        ("force_node_maps", 1 << 2),
        ),
    SInt16("unknown_0", VISIBLE=False),
    SInt32("model_checksum"),
    h3_reflexive("regions", mode_region),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    SInt32("instance_starting_mesh_index"),
    h3_reflexive("instances", mode_instance),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    h3_reflexive("nodes", mode_node),
    h3_reflexive("marker_groups", mode_marker_group),
    h3_reflexive("materials", mode_material),
    BytesRaw("unknown_3", SIZE=16, VISIBLE=False),
    SInt32("unknown_4", VISIBLE=False),
    h3_reflexive("meshes", mode_meshe),
    h3_reflexive("compression_info", mode_compression_info),
    h3_reflexive("unknown_nodey", mode_unknown_nodey),
    h3_reflexive("unknown_5", mode_unknown_5),
    BytesRaw("unknown_6", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_meshes", mode_unknown_meshe),
    h3_reflexive("node_maps", mode_node_map),
    h3_reflexive("unknown_7", mode_unknown_7),
    BytesRaw("unknown_8", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_yo", mode_unknown_yo),
    zone_asset("asset_data"),
    h3_reflexive("unknown_9", mode_unknown_9),
    BytesRaw("unknown_a", SIZE=192, VISIBLE=False),
    h3_reflexive("unknown_10", mode_unknown_10),
    h3_reflexive("runtime_nodes", mode_runtime_node),
    ENDIAN=">", SIZE=460
    )


def get():
    return mode_def

mode_def = TagDef("mode",
    h3_blam_header('mode'),
    mode_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["mode"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from struct import Struct as PyStruct
from traceback import format_exc

from reclaimer.model.constants import (
    JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN, SCALE_INTERNAL_TO_JMS )
from reclaimer.model.jms import write_jms, JmsModel, JmsNode,\
     JmsMaterial, JmsMarker, JmsVertex, JmsTriangle
from reclaimer.util.compression import decompress_normal32


__all__ = ("extract_model", )


def extract_model(tagdata, tag_path="", **kw):
    do_write_jms = kw.get('write_jms', True)
    if do_write_jms:
        jms_models = None
        filepath_base = Path(kw.get("out_dir", "")).joinpath(
            Path(tag_path).parent, "models")
    else:
        jms_models = []
        filepath_base = Path("")

    global_markers = {}
    materials = []
    regions = []
    nodes = []

    for b in tagdata.markers.STEPTREE:
        marker_name = b.name

        for inst in b.marker_instances.STEPTREE:
            try:
                region = tagdata.regions.STEPTREE[inst.region_index]
            except Exception:
                print("Invalid region index in marker '%s'" % marker_name)
                continue

            try:
                perm = region.permutations.STEPTREE[inst.permutation_index]
                perm_name = perm.name
                if (perm.flags.cannot_be_chosen_randomly and
                    not perm_name.startswith(JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN)):
                    perm_name = JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN + perm_name
            except Exception:
                print("Invalid permutation index in marker '%s'" % marker_name)
                continue

            perm_markers = global_markers.setdefault(perm_name, [])

            trans = inst.translation
            rot = inst.rotation
            perm_markers.append(JmsMarker(
                marker_name, perm_name, inst.region_index, inst.node_index,
                rot.i, rot.j, rot.k, rot.w,
                trans.x * SCALE_INTERNAL_TO_JMS,
                trans.y * SCALE_INTERNAL_TO_JMS,
                trans.z * SCALE_INTERNAL_TO_JMS,
                1.0
                ))

    for b in tagdata.nodes.STEPTREE:
        trans = b.translation
        rot = b.rotation
        nodes.append(JmsNode(
            b.name, b.first_child_node, b.next_sibling_node,
            rot.i, rot.j, rot.k, rot.w,
            trans.x * SCALE_INTERNAL_TO_JMS,
            trans.y * SCALE_INTERNAL_TO_JMS,
            trans.z * SCALE_INTERNAL_TO_JMS,
            b.parent_node
            ))

    for b in tagdata.shaders.STEPTREE:
        materials.append(JmsMaterial(
            b.shader.filepath.split("/")[-1].split("\\")[-1])
            )

    markers_by_perm = {}
    geoms_by_perm_lod_region = {}

    u_scale = tagdata.base_map_u_scale
    v_scale = tagdata.base_map_v_scale

    for region in tagdata.regions.STEPTREE:
        region_index = len(regions)
        regions.append(region.name)
        for perm in region.permutations.STEPTREE:
            perm_name = perm.name
            if (perm.flags.cannot_be_chosen_randomly and
                not perm_name.startswith(JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN)):
                perm_name = JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN + perm_name

            geoms_by_lod_region = geoms_by_perm_lod_region.setdefault(perm_name, {})

            perm_markers = markers_by_perm.setdefault(perm_name, [])
            if hasattr(perm, "local_markers"):
                for m in perm.local_markers.STEPTREE:
                    trans = m.translation
                    rot = m.rotation
                    perm_markers.append(JmsMarker(
                        m.name, perm_name, region_index, m.node_index,
                        rot.i, rot.j, rot.k, rot.w,
                        trans.x * SCALE_INTERNAL_TO_JMS,
                        trans.y * SCALE_INTERNAL_TO_JMS,
                        trans.z * SCALE_INTERNAL_TO_JMS,
                        1.0
                        ))

            last_geom_index = -1
            for lod in range(5):
                geoms_by_region = geoms_by_lod_region.get(lod, {})
                region_geoms = geoms_by_region.get(region.name, [])

                geom_index = perm[
                    perm.NAME_MAP["superlow_geometry_block"] + (4 - lod)]

                if (geom_index in region_geoms or
                    geom_index == last_geom_index):
                    continue

                geoms_by_lod_region[lod] = geoms_by_region
                geoms_by_region[region.name] = region_geoms
                region_geoms.append(geom_index)
                last_geom_index = geom_index

    try:
        use_local_nodes = tagdata.flags.parts_have_local_nodes
    except Exception:
        use_local_nodes = False
    def_node_map = list(range(128))
    def_node_map.append(-1)

    # use big endian since it will have been byteswapped
    comp_vert_unpacker = PyStruct(">3f3I2h2bh").unpack_from
    uncomp_vert_unpacker = PyStruct(">14f2h2f").unpack_from

    for perm_name in sorted(geoms_by_perm_lod_region):
        geoms_by_lod_region = geoms_by_perm_lod_region[perm_name]
        perm_markers = markers_by_perm.get(perm_name)

        for lod in sorted(geoms_by_lod_region):
            if lod == -1:
                continue

            jms_name = perm_name + {4: " superlow", 3: " low", 2: " medium",
                                    1: " high", 0: " superhigh"}.get(lod, "")

            filepath = filepath_base.joinpath(jms_name + ".jms")

            markers = list(perm_markers)
            markers.extend(global_markers.get(perm_name, ()))
            verts = []
            tris = []

            geoms_by_region = geoms_by_lod_region[lod]
            for region_name in sorted(geoms_by_region):
                region_index = regions.index(region_name)
                geoms = geoms_by_region[region_name]

                for geom_index in geoms:
                    try:
                        geom_block = tagdata.geometries.STEPTREE[geom_index]
                    except Exception:
                        print("Invalid geometry index '%s'" % geom_index)
                        continue

                    for part in geom_block.parts.STEPTREE:
                        v_origin = len(verts)
                        shader_index = part.shader_index

                        try:
                            node_map = list(part.local_nodes)
                            node_map.append(-1)
                            compressed = False
                        except (AttributeError, KeyError):
                            compressed = True

                        if not use_local_nodes:
                            node_map = def_node_map

                        try:
                            unparsed = isinstance(
                                part.triangles.STEPTREE.data, bytearray)
                        except Exception:
                            unparsed = False

                        # TODO: Make this work in meta(parse verts and tris)
                        try:
                            if compressed and unparsed:
                                vert_data = part.compressed_vertices.STEPTREE.data
                                for off in range(0, len(vert_data), 32):
                                    v = comp_vert_unpacker(vert_data, off)
                                    verts.append(JmsVertex(
                                        v[8]//3,
                                        v[0] * 100, v[1] * 100, v[2] * 100,
                                        *decompress_normal32(v[3]),
                                        v[9]//3, 1.0 - (v[10]/32767),
                                        u_scale * v[6]/32767, 1.0 - v_scale * v[7]/32767))
                            elif compressed:
                                for v in part.compressed_vertices.STEPTREE:
                                    verts.append(JmsVertex(
                                        v[8]//3,
                                        v[0] * 100, v[1] * 100, v[2] * 100,
                                        *decompress_normal32(v[3]),
                                        v[9]//3, 1.0 - (v[10]/32767),
                                        u_scale * v[6]/32767, 1.0 - v_scale * v[7]/32767))
                            elif not compressed and unparsed:
                                vert_data = part.uncompressed_vertices.STEPTREE.data
                                for off in range(0, len(vert_data), 68):
                                    v = uncomp_vert_unpacker(vert_data, off)
                                    verts.append(JmsVertex(
                                        node_map[v[14]],
                                        v[0] * 100, v[1] * 100, v[2] * 100,
                                        v[3], v[4], v[5],
                                        node_map[v[15]], max(0, min(1, v[17])),
                                        u_scale * v[12], 1.0 - v_scale * v[13]))
                            else:
                                for v in part.uncompressed_vertices.STEPTREE:
                                    verts.append(JmsVertex(
                                        node_map[v[14]],
                                        v[0] * 100, v[1] * 100, v[2] * 100,
                                        v[3], v[4], v[5],
                                        node_map[v[15]], max(0, min(1, v[17])),
                                        u_scale * v[12], 1.0 - v_scale * v[13]))
                        except Exception:
                            print(format_exc())
                            print("If you see this, tell Moses to stop "
                                  "fucking with the vertex definition.")

                        try:
                            if unparsed:
                                tri_block = part.triangles.STEPTREE.data
                                tri_list = [-1] * (len(tri_block) // 2)
                                for i in range(len(tri_list)):
                                    # assuming big endian
                                    tri_list[i] = (
                                        tri_block[i * 2 + 1] +
                                        (tri_block[i * 2] << 8))
                                    if tri_list[i] > 32767:
                                        tri_list[i] = -1
                            else:
                                tri_block = part.triangles.STEPTREE
                                tri_list = []
                                for triangle in tri_block:
                                    tri_list.extend(triangle)

                            swap = True
                            for i in range(len(tri_list) - 2):
                                v0 = tri_list[i]
                                v1 = tri_list[i + 1 + swap]
                                v2 = tri_list[i + 2 - swap]
                                if v0 != -1 and v1 != -1 and v2 != -1:
                                    # remove degens
                                    if v0 != v1 and v0 != v2 and v1 != v2:
                                        tris.append(JmsTriangle(
                                            region_index, shader_index,
                                            v0 + v_origin,
                                            v1 + v_origin,
                                            v2 + v_origin))
                                swap = not swap
                        except Exception:
                            print(format_exc())
                            print("Could not parse triangle blocks.")

            jms_model = JmsModel(
                jms_name, tagdata.node_list_checksum, nodes,
                materials, markers, regions, verts, tris)
            if do_write_jms:
                write_jms(filepath, jms_model)
            else:
                jms_models.append(jms_model)

    return jms_models



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .objs.magy import MagyTag
from ...hek.defs.antr import *

magy_body = Struct("tagdata",
    reflexive("objects",  object_desc, 4),
    reflexive("units",    unit_desc, 32, DYN_NAME_PATH=".label"),
    reflexive("weapons",  weapon_desc, 1),
    reflexive("vehicles", vehicle_desc, 1),
    reflexive("devices",  device_desc, 1),
    reflexive("unit_damages", anim_enum_desc, 176,
        *unit_damage_animation_names
        ),
    reflexive("fp_animations", fp_animation_desc, 1),
    #i have no idea why they decided to cap it at 257 instead of 256....
    reflexive("sound_references", sound_reference_desc, 257,
        DYN_NAME_PATH=".sound.filepath"),
    Float("limp_body_node_radius"),
    Bool16("flags",
        "compress_all_animations",
        "force_idle_compression",
        ),
    Pad(2),
    reflexive("nodes", nodes_desc, 64, DYN_NAME_PATH=".name"),
    reflexive("animations", animation_desc, 2048, DYN_NAME_PATH=".name"),
    dependency_os("stock_animation", valid_model_animations_yelo),
    SIZE=300,
    )


def get():
    return magy_def

magy_def = TagDef("magy",
    blam_header_os('magy', 0),
    magy_body,

    ext=".model_animations_yelo", endian=">", tag_cls=MagyTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import traceback

from copy import deepcopy
from math import sqrt
from pathlib import Path

from reclaimer.util.matrices import axis_angle_to_quat, multiply_quaternions
from reclaimer.animation.jma import JmsNode, JmaAnimation, JmaRootNodeState,\
     JmaNodeState, write_jma, get_anim_ext
from reclaimer.animation import serialization
from reclaimer.model import jms

__all__ = ("extract_model_animations", "extract_animation", )


def antr_nodes_to_jms_nodes(anim_nodes):
    nodes = []
    for anim_node in anim_nodes:
        nodes.append(
            JmsNode(anim_node.name, anim_node.first_child_node_index,
                    anim_node.next_sibling_node_index,
                    parent_index=anim_node.parent_node_index)
            )

    return nodes


def extract_model_animations(tagdata, tag_path="", **kw):
    return_jma = not kw.get('write_jma', True)
    jma_anims = [] if return_jma else None
    if not tagdata.animations.STEPTREE:
        return jma_anims

    for i in range(len(tagdata.animations.STEPTREE)):
        try:
            result = extract_animation(i, tagdata, tag_path, **kw)
            if result and return_jma:
                jma_anims.append(result)
        except Exception:
            print(traceback.format_exc())
            print("Could not extract animation.")

    return jma_anims


def extract_animation(anim_index, tagdata, tag_path="", **kw):
    endian = kw.get("endian", ">")

    anim_nodes = antr_nodes_to_jms_nodes(tagdata.nodes.STEPTREE)
    anim = tagdata.animations.STEPTREE[anim_index]

    anim_ext = get_anim_ext(anim.type.enum_name,
                            anim.frame_info_type.enum_name,
                            anim.flags.world_relative)

    filepath = Path("")
    do_write_jma = kw.get('write_jma', True)
    if do_write_jma:
        filepath = Path(kw.get("out_dir", "")).joinpath(
            Path(tag_path).parent, "animations", anim.name + anim_ext)
        if not kw.get('overwrite', True) and filepath.is_file():
            return

    if len(anim_nodes) != anim.node_count:
        print("WARNING: Animation tag missing nodes.\n"
              "\tFake nodes will be created to allow compiling the animations.\n"
              "\tAnimation tags compiled from these files won't import onto\n"
              "\ttheir gbxmodel in 3DS Max, as their node names won't match.")
        anim_nodes = jms.util.generate_fake_nodes(anim.node_count)

    if len(anim_nodes) != anim.node_count:
        print("Skipping animation with different number of nodes "
              "than the tag contains: '%s'" % anim.name)
        return


    jma_anim = JmaAnimation(
        anim.name, anim.node_list_checksum, anim.type.enum_name,
        anim.frame_info_type.enum_name, anim.flags.world_relative,
        anim_nodes
        )
    jma_anim.trans_flags_int = anim.trans_flags0 | (anim.trans_flags1 << 32)
    jma_anim.rot_flags_int   = anim.rot_flags0   | (anim.rot_flags1   << 32)
    jma_anim.scale_flags_int = anim.scale_flags0 | (anim.scale_flags1 << 32)

    if len(anim.frame_info.STEPTREE) < jma_anim.root_node_info_frame_size * anim.frame_count:
        print("Skipping animation with less frame_info data "
              "than it is expected to contain: '%s'" % anim.name)
        return
    elif len(anim.default_data.STEPTREE) < jma_anim.default_data_size:
        print("Skipping animation with less default_data "
              "than it is expected to contain: '%s'" % anim.name)
        return
    elif len(anim.frame_data.STEPTREE) < jma_anim.frame_data_frame_size * anim.frame_count:
        print("Skipping animation with less frame_data "
              "than it is expected to contain: '%s'" % anim.name)
        return

    if anim.flags.compressed_data:
        # decompress compressed animations
        keyframes, jma_anim.frames = serialization.deserialize_compressed_frame_data(anim)
        jma_anim.rot_keyframes   = keyframes[0]
        jma_anim.trans_keyframes = keyframes[1]
        jma_anim.scale_keyframes = keyframes[2]
    else:
        # create the node states from the frame_data and default_data
        jma_anim.frames = serialization.deserialize_frame_data(
            anim, None, True, endian)

    # sum the frame info changes for each frame from the frame_info
    jma_anim.root_node_info = serialization.deserialize_frame_info(
        anim, True, endian)

    if jma_anim.has_dxdy or jma_anim.has_dyaw or jma_anim.has_dz:
        # this is set to True on instantiation.
        # Set it to False since we had to provide root node info
        jma_anim.root_node_info_applied = False
        jma_anim.apply_root_node_info_to_states()

    if do_write_jma:
        write_jma(filepath, jma_anim)
    else:
        return jma_anim



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

smet_attrs = Struct("smet_attrs",
    #Meter Shader Properties
    Struct("meter_shader",
        Bool16("meter_shader_flags",
            "decal",
            "two_sided",
            "flash_color_is_negative",
            "tint_mode_2",
            "unfiltered"
            ),
        Pad(34),
        dependency("map", "bitm"),
        Pad(32),
        ),

    #Colors
    Struct("colors",
        Struct("gadient_min", INCLUDE=rgb_float),
        Struct("gadient_max", INCLUDE=rgb_float),
        Struct("background", INCLUDE=rgb_float),
        Struct("flash", INCLUDE=rgb_float),
        Struct("tint", INCLUDE=rgb_float),
        float_zero_to_one("meter_transparency"),
        float_zero_to_one("background_transparency"),
        ),

    Pad(24),
    #External Function Sources
    Struct("external_function_sources",
        SEnum16("meter_brightness", *function_outputs),
        SEnum16("flash_brightness", *function_outputs),
        SEnum16("value",            *function_outputs),
        SEnum16("gradient",         *function_outputs),
        SEnum16("flash_extension",  *function_outputs),
        ),
    SIZE=220,
    )

smet_body = Struct("tagdata",
    shdr_attrs,
    smet_attrs,
    SIZE=260,
    )



def get():
    return smet_def

smet_def = TagDef("smet",
    blam_header('smet'),
    smet_body,

    ext=".shader_transparent_meter", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .yelo import *
from ...hek.defs.objs.tag import HekTag

gelo_body = Struct("tagdata",
    SInt16("version"),
    Bool16("flags",
        "hide_health_when_zoomed",
        "hide_shield_when_zoomed",
        "hide_motion_sensor_when_zoomed",
        "force_game_to_use_stun_jumping_penalty"
        ),
    SInt32("base_address"),
    ascii_str32("mod_name"),
    dependency_os("global_explicit_references", "tagc"),
    dependency_os("chokin_victim_globals", "gelc"),

    Pad(16),
    Pad(12), #reflexive("unknown1", void_desc),
    Pad(52),
    reflexive("scripted_ui_widgets", scripted_ui_widget, 128),

    Pad(12), #reflexive("unknown2", void_desc),
    Pad(20),
    reflexive("yelo_scripting", yelo_scripting, 1),

    SIZE=288
    )

def get():
    return gelo_def

gelo_def = TagDef("gelo",
    blam_header_os('gelo', 2),
    gelo_body,

    ext=".project_yellow_globals", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.matg import *
from supyr_struct.defs.tag_def import TagDef
from ..common_descs import *

def get():
    return matg_def

matg_body = Struct('tagdata',
    Pad(248),
    reflexive("sounds", sound, 2,
        "enter water", "exit water"),
    reflexive("cameras", camera, 1),
    reflexive("player_controls", player_control, 1),
    reflexive("difficulties", difficulty, 1),
    reflexive("grenades", grenade, len(grenade_types), *grenade_types),
    reflexive("rasterizer_datas", rasterizer_data, 1),
    reflexive("interface_bitmaps", interface_bitmaps, 1),
    reflexive("cheat_weapons", cheat_weapon, 20,
        DYN_NAME_PATH='.weapon.filepath'),
    reflexive("cheat_powerups", cheat_powerup, 20,
        DYN_NAME_PATH='.powerup.filepath'),
    reflexive("multiplayer_informations", multiplayer_information, 1),
    reflexive("player_informations", player_information, 1),
    reflexive("first_person_interfaces", first_person_interface, 1),
    reflexive("falling_damages", falling_damage, 1),
    reflexive("materials", material, len(materials_list), *materials_list),
    reflexive("playlist_members", playlist_member, 20,
        DYN_NAME_PATH='.map_name'),

    SIZE=428
    )

matg_def = TagDef("matg",
    blam_header_stubbs('matg', 3),
    matg_body,

    ext=".globals", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Updated and copypasted.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


bkey_text_widget = Struct("text_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("left_alignment", 1 << 3),
        "right_alignment",
        ("all_caps", 1 << 6),
        ("wrap_at_bounds", 1 << 13),
        ),
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_string_id("data_source_name"),
    h3_string_id("text_string"),
    h3_string_id("text_color"),
    SInt16("text_font"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=76
    )


bkey_bitmap_widget = Struct("bitmap_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("scale_to_bounds", 1 << 3),
        "replace_with_blur",
        ("replace_with_white", 1 << 10),
        "replace_with_black",
        ),
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_dependency("bitmap"),
    h3_dependency("unknown_1"),
    SEnum16("blend_method", *bmp3_blend_method),
    SInt16("unknown_2", VISIBLE=False),
    SInt16("sprite_index"),
    SInt16("unknown_3", VISIBLE=False),
    h3_string_id("data_source_name"),
    h3_string_id("sprite_data_source_name"),
    ENDIAN=">", SIZE=108
    )


bkey_body = Struct("tagdata",
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_dependency("strings"),
    h3_reflexive("text_widgets", bkey_text_widget),
    h3_reflexive("bitmap_widgets", bkey_bitmap_widget),
    ENDIAN=">", SIZE=84
    )


def get():
    return bkey_def

bkey_def = TagDef("bkey",
    h3_blam_header('bkey'),
    bkey_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["bkey"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from supyr_struct.defs.constants import DEFAULT
from supyr_struct.tag import Tag

class HekTag(Tag):
    def __init__(self, **kwargs):
        self.calc_pointers = False
        Tag.__init__(self, **kwargs)

    def calc_internal_data(self):
        # recalculate the header data
        head = self.data.blam_header

        head.tag_class.data = head.tag_class.get_desc(DEFAULT)
        head.checksum = head.get_desc(DEFAULT, 'checksum')
        head.header_size = head.get_desc(DEFAULT, 'header_size')
        head.version = head.get_desc(DEFAULT, 'version')
        head.integrity0 = head.get_desc(DEFAULT, 'integrity0')
        head.integrity1 = head.get_desc(DEFAULT, 'integrity1')
        head.engine_id.data = head.engine_id.get_desc(DEFAULT)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    import arbytmap
    if not hasattr(arbytmap, "FORMAT_P8"):
        arbytmap.FORMAT_P8 = "P8"

        """ADD THE P8 FORMAT TO THE BITMAP CONVERTER"""
        arbytmap.register_format(format_id=arbytmap.FORMAT_P8,
                                 depths=(8,8,8,8))

    if not hasattr(arbytmap, "FORMAT_P8_BUMP"):
        arbytmap.FORMAT_P8_BUMP = "P8-BUMP"

        """ADD THE P8 FORMAT TO THE BITMAP CONVERTER"""
        arbytmap.register_format(format_id=arbytmap.FORMAT_P8_BUMP,
                                 depths=(8,8,8,8))

    from arbytmap import Arbytmap, bitmap_io, TYPE_2D, TYPE_3D, TYPE_CUBEMAP,\
         FORMAT_A8, FORMAT_L8, FORMAT_AL8, FORMAT_A8L8,\
         FORMAT_R5G6B5, FORMAT_A1R5G5B5, FORMAT_A4R4G4B4,\
         FORMAT_X8R8G8B8, FORMAT_A8R8G8B8,\
         FORMAT_DXT1, FORMAT_DXT3, FORMAT_DXT5, FORMAT_CTX1, FORMAT_DXN,\
         FORMAT_DXT3Y, FORMAT_DXT3A, FORMAT_DXT3AY,\
         FORMAT_DXT5Y, FORMAT_DXT5A, FORMAT_DXT5AY,\
         FORMAT_P8_BUMP, FORMAT_P8, FORMAT_V8U8, FORMAT_R8G8,\
         FORMAT_R16G16B16F, FORMAT_A16R16G16B16F,\
         FORMAT_R32G32B32F, FORMAT_A32R32G32B32F
except ImportError:
    arbytmap = Arbytmap = None

import zlib

from array import array
from pathlib import Path
from struct import unpack

from reclaimer.util import get_is_xbox_map
from reclaimer.h3.util import get_virtual_dimension,\
     get_pixel_bytes_size, get_h3_pixel_bytes_size
from reclaimer.bitmaps.p8_palette import HALO_P8_PALETTE, STUBBS_P8_PALETTE

__all__ = ("extract_bitmap_tiff_data", "extract_bitmaps", )


#each sub-bitmap(cubemap face) must be a multiple of 128 bytes
CUBEMAP_PADDING = 128


def extract_bitmap_tiff_data(tag_path):
    tag_path = Path(tag_path)
    try:
        with tag_path.open('rb') as f:
            tag_header_data = f.read(256)  # 256 is enough to read all the
            #                                pertinent header information

        tag_id = tag_header_data[36: 40]
        engine_id = tag_header_data[60: 64]

        # make sure this is a bitmap tag
        if tag_id == b'bitm' and engine_id == b'blam':
            # halo 1
            dims_off = 64+24
            size_off = 64+28
            data_off = 64+108
            endian = ">"
        elif tag_id == b'mtib' and engine_id == b'!MLB':
            # halo 2
            dims_off = 64+16+24
            size_off = 64+16+28
            data_off = 64+16
            # get the size of the bitmap body from the tbfd structure
            data_off += unpack("<i", tag_header_data[data_off-4: data_off])[0]
            endian = "<"
        else:
            print('    This file doesnt appear to be a bitmap tag.')
            return 0, 0, bytearray()

        width, height = unpack(endian + "HH", tag_header_data[dims_off: dims_off+4])
        comp_size = unpack(endian + "i", tag_header_data[size_off: size_off+4])[0]
    except Exception:
        print('    Could not load bitmap tag.')
        return 0, 0, bytearray()


    with tag_path.open('rb') as f:
        f.seek(data_off)
        comp_data = f.read(comp_size)

    if not comp_data:
        print('    No source image to extract.')
        return width, height, bytearray()

    try:
        data_size = unpack(endian + "I", comp_data[:4])[0]
        if not data_size:
            print('    Source data is blank.')
            return width, height, bytearray()

        return width, height, bytearray(zlib.decompress(comp_data[4:]))
    except Exception:
        print('    Could not decompress data.')
        return width, height, bytearray()


def extract_bitmaps(tagdata, tag_path, **kw):
    out_dir = Path(kw.get("out_dir", ""))
    filepath_base = out_dir.joinpath(tag_path).parent
    filename_base = Path(tag_path).name

    ext = kw.get("bitmap_ext", "").strip(". ")
    keep_alpha = kw.get("bitmap_keep_alpha", True)
    engine = kw.pop('engine', '')
    pix_data = tagdata.processed_pixel_data.STEPTREE
    if 'halo_map' in kw:
        engine = kw['halo_map'].engine

    p8_palette = STUBBS_P8_PALETTE if "stubbs" in engine else HALO_P8_PALETTE
    if not ext:
        ext = "dds"

    is_xbox = get_is_xbox_map(engine)
    is_gen3 = hasattr(tagdata, "zone_assets_normal")
    if Arbytmap is None:
        # cant extract xbox bitmaps yet
        return "    Arbytmap not loaded. Cannot extract bitmaps."

    arby = Arbytmap()
    bitm_i = 0
    multi_bitmap = len(tagdata.bitmaps.STEPTREE) > 1
    size_calc = get_h3_pixel_bytes_size if is_gen3 else get_pixel_bytes_size
    dim_calc = get_virtual_dimension if is_gen3 else None

    for bitmap in tagdata.bitmaps.STEPTREE:
        typ = bitmap.type.enum_name
        fmt = bitmap.format.enum_name
        w = bitmap.width
        h = bitmap.height
        d = bitmap.depth
        tiled = False
        if hasattr(bitmap, "format_flags"):
            tiled = bitmap.format_flags.tiled

        filename = filename_base
        if multi_bitmap:
            filename += "__%s" % bitm_i
            bitm_i += 1

        tex_block = []
        tex_info = dict(
            width=w, height=h, depth=d, mipmap_count=bitmap.mipmaps,
            swizzled=bitmap.flags.swizzled, big_endian=is_gen3,
            packed=True, tiled=tiled, tile_method="DXGI",
            packed_width_calc=dim_calc, packed_height_calc=dim_calc,
            filepath=str(filepath_base.joinpath(filename + "." + ext))
            )
        tex_info["texture_type"] = {
            "texture_2d": TYPE_2D, "texture_3d": TYPE_3D,
            "cubemap": TYPE_CUBEMAP}.get(typ, TYPE_2D)
        tex_info["sub_bitmap_count"] = {
            "texture_2d": 1, "texture_3d": 1,
            "cubemap": 6, "multipage_2d": d}.get(typ, 1)
        if typ == "multipage_2d":
            tex_info.update(depth=1)
            d = 1


        if fmt == "p8_bump":
            tex_info.update(
                palette=[p8_palette.p8_palette_32bit_packed]*(bitmap.mipmaps + 1),
                palette_packed=True, indexing_size=8, format=FORMAT_P8_BUMP)
        else:
            tex_info["format"] = {
                "a8": FORMAT_A8, "y8": FORMAT_L8, "ay8": FORMAT_AL8,
                "a8y8": FORMAT_A8L8, "p8": FORMAT_A8,
                "v8u8": FORMAT_V8U8, "g8b8": FORMAT_R8G8,
                "x8r8g8b8": FORMAT_A8R8G8B8, "a8r8g8b8": FORMAT_A8R8G8B8,
                "r5g6b5": FORMAT_R5G6B5, "a1r5g5b5": FORMAT_A1R5G5B5,
                "a4r4g4b4": FORMAT_A4R4G4B4,
                "dxt1": FORMAT_DXT1, "dxt3": FORMAT_DXT3, "dxt5": FORMAT_DXT5,
                "ctx1": FORMAT_CTX1, "dxn": FORMAT_DXN, "dxt5ay": FORMAT_DXT5AY,
                "dxt3a": FORMAT_DXT3A, "dxt3y": FORMAT_DXT3Y,
                "dxt5a": FORMAT_DXT5A, "dxt5y": FORMAT_DXT5Y,
                "rgbfp16": FORMAT_R16G16B16F, "argbfp32": FORMAT_A32R32G32B32F,
                "rgbfp32": FORMAT_R32G32B32F}.get(fmt, None)


        arby_fmt = tex_info["format"]
        if arby_fmt is None:
            continue

        i_max = tex_info["sub_bitmap_count"] if is_xbox else bitmap.mipmaps + 1
        j_max = bitmap.mipmaps + 1 if is_xbox else tex_info['sub_bitmap_count']
        off = bitmap.pixels_offset
        for i in range(i_max):
            if not is_xbox:
                mip_size = size_calc(arby_fmt, w, h, d, i, tiled)

            for j in range(j_max):
                if is_xbox:
                    mip_size = size_calc(arby_fmt, w, h, d, j, tiled)

                if fmt == "p8_bump":
                    tex_block.append(
                        array('B', pix_data[off: off + (mip_size // 4)]))
                    off += len(tex_block[-1])
                else:
                    off = bitmap_io.bitmap_bytes_to_array(
                        pix_data, off, tex_block,
                        arby_fmt, 1, 1, 1, mip_size)

            # skip the xbox alignment padding to get to the next texture
            if is_xbox and typ == "cubemap":
                off += ((CUBEMAP_PADDING - (off % CUBEMAP_PADDING)) %
                        CUBEMAP_PADDING)


        if is_xbox and typ == "cubemap":
            template = tuple(tex_block)
            i = 0
            for f in (0, 2, 1, 3, 4, 5):
                for m in range(bitmap.mipmaps + 1):
                    tex_block[m*6 + f] = template[i]
                    i += 1

        if not tex_block:
            # nothing to extract
            continue

        arby.load_new_texture(texture_block=tex_block, texture_info=tex_info,
                              tile_mode=False, swizzle_mode=False)
        arby.save_to_file(keep_alpha=keep_alpha)
